@@ -298,7 +298,7 @@ int cmd_format(int argc, char *argv[])
 		struct printbuf buf = PRINTBUF;
 		buf.human_readable_units = true;
 
-		bch2_sb_to_text_with_names(&buf, sb, false, 1 << BCH_SB_FIELD_members_v2, -1);
+		bch2_sb_to_text_with_names(&buf, NULL, sb, false, 1 << BCH_SB_FIELD_members_v2, -1);
 		printf("%s", buf.buf);
 		printbuf_exit(&buf);
 	}
@@ -323,7 +323,7 @@ int cmd_format(int argc, char *argv[])
 		if (opts.source)
 			build_fs(c, opts.source);
 
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 	}
 	bch2_opt_strs_free(&fs_opt_strs);
 	darray_exit(&devices);
