@@ -110,24 +110,72 @@ static int find_snapshot_tree_subvol(struct btree_trans *trans,
 	return ret ?: bch_err_throw(trans->c, ENOENT_no_snapshot_tree_subvol);
 }
 
+static struct qstr lostfound_str = QSTR("lost+found");
+
+static int create_lostfound(struct btree_trans *trans, u32 snapshot_tree,
+			    subvol_inum root_inum,
+			    struct bch_inode_unpacked *root_inode,
+			    struct bch_hash_info *root_hash_info,
+			    struct bch_inode_unpacked *lostfound)
+{
+	struct bch_fs *c = trans->c;
+	/*
+	 * We always create lost+found in the root snapshot; we don't want
+	 * different branches of the snapshot tree to have different lost+found
+	 */
+	struct bch_snapshot_tree st;
+	try(bch2_snapshot_tree_lookup(trans, snapshot_tree, &st));
+
+	u32 snapshot = bch2_snapshot_live_descendent(c, le32_to_cpu(st.root_snapshot));
+
+	CLASS(bch_log_msg_level, msg)(c, LOGLEVEL_notice);
+	prt_printf(&msg.m, "creating ");
+	try(bch2_inum_to_path(trans, root_inum, &msg.m));
+	prt_printf(&msg.m, "/lost+found in subvol %llu snapshot %u", root_inum.subvol, snapshot);
+
+	u64 now = bch2_current_time(c);
+	u64 cpu = raw_smp_processor_id();
+
+	bch2_inode_init_early(c, lostfound);
+	bch2_inode_init_late(c, lostfound, now, 0, 0, S_IFDIR|0700, 0, root_inode);
+	lostfound->bi_dir = root_inode->bi_inum;
+	lostfound->bi_snapshot = snapshot;
+
+	root_inode->bi_nlink++;
+
+	CLASS(btree_iter_uninit, lostfound_iter)(trans);
+	try(bch2_inode_create(trans, &lostfound_iter, lostfound, snapshot, cpu,
+			      inode_opt_get(c, root_inode, inodes_32bit)));
+
+	bch2_btree_iter_set_snapshot(&lostfound_iter, snapshot);
+	try(bch2_btree_iter_traverse(&lostfound_iter));
+
+	try(bch2_dirent_create_snapshot(trans,
+				0, root_inode->bi_inum, snapshot, root_hash_info,
+				mode_to_type(lostfound->bi_mode),
+				&lostfound_str,
+				lostfound->bi_inum,
+				&lostfound->bi_dir_offset,
+				BTREE_UPDATE_internal_snapshot_node|
+				STR_HASH_must_create));
+
+	try(bch2_inode_write_flags(trans, &lostfound_iter, lostfound,
+				   BTREE_UPDATE_internal_snapshot_node));
+
+	return bch2_trans_commit_lazy(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc);
+}
+
 /* Get lost+found, create if it doesn't exist: */
 static int lookup_lostfound(struct btree_trans *trans, u32 snapshot,
 			    struct bch_inode_unpacked *lostfound,
 			    u64 reattaching_inum)
 {
 	struct bch_fs *c = trans->c;
-	struct qstr lostfound_str = QSTR("lost+found");
-	CLASS(btree_iter_uninit, lostfound_iter)(trans);
-	u64 inum = 0;
-	unsigned d_type = 0;
+	u32 snapshot_tree = bch2_snapshot_tree(c, snapshot);
 	int ret;
 
-	struct bch_snapshot_tree st;
-	try(bch2_snapshot_tree_lookup(trans, bch2_snapshot_tree(c, snapshot), &st));
-
 	u32 subvolid;
-	ret = find_snapshot_tree_subvol(trans,
-				bch2_snapshot_tree(c, snapshot), &subvolid);
+	ret = find_snapshot_tree_subvol(trans, snapshot_tree, &subvolid);
 	bch_err_msg(c, ret, "finding subvol associated with snapshot tree %u",
 		    bch2_snapshot_tree(c, snapshot));
 	if (ret)
@@ -153,27 +201,38 @@ static int lookup_lostfound(struct btree_trans *trans, u32 snapshot,
 	};
 
 	struct bch_inode_unpacked root_inode;
-	struct bch_hash_info root_hash_info;
 	ret = bch2_inode_find_by_inum_snapshot(trans, root_inum.inum, snapshot, &root_inode, 0);
 	bch_err_msg(c, ret, "looking up root inode %llu for subvol %u",
 		    root_inum.inum, subvolid);
 	if (ret)
 		return ret;
 
-	root_hash_info = bch2_hash_info_init(c, &root_inode);
+	struct bch_hash_info root_hash_info;
+	try(bch2_hash_info_init(c, &root_inode, &root_hash_info));
 
+	u64 inum = 0;
+	unsigned d_type = 0;
 	ret = lookup_dirent_in_snapshot(trans, root_hash_info, root_inum,
 			      &lostfound_str, &inum, &d_type, snapshot);
-	if (bch2_err_matches(ret, ENOENT))
-		goto create_lostfound;
+	if (bch2_err_matches(ret, ENOENT)) {
+		/*
+		 * We always create lost_found in its own transaction; this will
+		 * return a transaction restart:
+		 */
+		ret = create_lostfound(trans, snapshot_tree, root_inum,
+				       &root_inode, &root_hash_info, lostfound);
+		bch_err_msg(c, ret, "creating lost+found");
+		return ret;
+	}
 
 	bch_err_fn(c, ret);
 	if (ret)
 		return ret;
 
 	if (d_type != DT_DIR) {
-		bch_err(c, "error looking up lost+found: not a directory");
-		return bch_err_throw(c, ENOENT_not_directory);
+		ret = bch_err_throw(c, ENOENT_not_directory);
+		bch_err_msg(c, ret, "looking up lost+found");
+		return ret;
 	}
 
 	/*
@@ -184,58 +243,6 @@ static int lookup_lostfound(struct btree_trans *trans, u32 snapshot,
 	bch_err_msg(c, ret, "looking up lost+found %llu:%u in (root inode %llu, snapshot root %u)",
 		    inum, snapshot, root_inum.inum, bch2_snapshot_root(c, snapshot));
 	return ret;
-
-create_lostfound:
-	/*
-	 * we always create lost+found in the root snapshot; we don't want
-	 * different branches of the snapshot tree to have different lost+found
-	 */
-	snapshot = le32_to_cpu(st.root_snapshot);
-	/*
-	 * XXX: we could have a nicer log message here  if we had a nice way to
-	 * walk backpointers to print a path
-	 */
-	CLASS(printbuf, path)();
-	ret = bch2_inum_to_path(trans, root_inum, &path);
-	if (ret)
-		goto err;
-
-	bch_notice(c, "creating %s/lost+found in subvol %llu snapshot %u",
-		   path.buf, root_inum.subvol, snapshot);
-
-	u64 now = bch2_current_time(c);
-	u64 cpu = raw_smp_processor_id();
-
-	bch2_inode_init_early(c, lostfound);
-	bch2_inode_init_late(c, lostfound, now, 0, 0, S_IFDIR|0700, 0, &root_inode);
-	lostfound->bi_dir = root_inode.bi_inum;
-	lostfound->bi_snapshot = le32_to_cpu(st.root_snapshot);
-
-	root_inode.bi_nlink++;
-
-	ret = bch2_inode_create(trans, &lostfound_iter, lostfound, snapshot, cpu,
-				inode_opt_get(c, &root_inode, inodes_32bit));
-	if (ret)
-		goto err;
-
-	bch2_btree_iter_set_snapshot(&lostfound_iter, snapshot);
-	ret = bch2_btree_iter_traverse(&lostfound_iter);
-	if (ret)
-		goto err;
-
-	ret =   bch2_dirent_create_snapshot(trans,
-				0, root_inode.bi_inum, snapshot, &root_hash_info,
-				mode_to_type(lostfound->bi_mode),
-				&lostfound_str,
-				lostfound->bi_inum,
-				&lostfound->bi_dir_offset,
-				BTREE_UPDATE_internal_snapshot_node|
-				STR_HASH_must_create) ?:
-		bch2_inode_write_flags(trans, &lostfound_iter, lostfound,
-				       BTREE_UPDATE_internal_snapshot_node);
-err:
-	bch_err_msg(c, ret, "creating lost+found");
-	return ret;
 }
 
 static inline bool inode_should_reattach(struct bch_inode_unpacked *inode)
@@ -337,7 +344,8 @@ int bch2_reattach_inode(struct btree_trans *trans, struct bch_inode_unpacked *in
 
 	try(__bch2_fsck_write_inode(trans, &lostfound));
 
-	struct bch_hash_info dir_hash = bch2_hash_info_init(c, &lostfound);
+	struct bch_hash_info dir_hash;
+	try(bch2_hash_info_init(c, &lostfound, &dir_hash));
 	struct qstr name = QSTR(name_buf);
 
 	inode->bi_dir = lostfound.bi_inum;
@@ -913,6 +921,9 @@ static int check_inode(struct btree_trans *trans,
 	}
 
 	ret = bch2_check_inode_has_case_insensitive(trans, &u, &s->ids, &do_update);
+	if (bch2_err_matches(ret, ENOENT)) /* disconnected inode; will be fixed by a later pass */
+		ret = 0;
+	bch_err_msg(c, ret, "bch2_check_inode_has_case_insensitive()");
 	if (ret)
 		goto err;
 
@@ -1089,14 +1100,14 @@ int bch2_check_inodes(struct bch_fs *c)
 	CLASS(btree_trans, trans)(c);
 	CLASS(snapshots_seen, s)();
 
-	struct progress_indicator_state progress;
-	bch2_progress_init(&progress, c, BIT_ULL(BTREE_ID_inodes));
+	struct progress_indicator progress;
+	bch2_progress_init(&progress, __func__, c, BIT_ULL(BTREE_ID_inodes), 0);
 
 	return for_each_btree_key_commit(trans, iter, BTREE_ID_inodes,
 				POS_MIN,
 				BTREE_ITER_prefetch|BTREE_ITER_all_snapshots, k,
 				NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter) ?:
+		bch2_progress_update_iter(trans, &progress, &iter) ?:
 		check_inode(trans, &iter, k, &snapshot_root, &s);
 	}));
 }
@@ -1127,9 +1138,7 @@ static int find_oldest_inode_needs_reattach(struct btree_trans *trans,
 			break;
 
 		struct bch_inode_unpacked parent_inode;
-		ret = bch2_inode_unpack(k, &parent_inode);
-		if (ret)
-			break;
+		try(bch2_inode_unpack(k, &parent_inode));
 
 		if (!inode_should_reattach(&parent_inode))
 			break;
@@ -1162,7 +1171,7 @@ static int check_unreachable_inode(struct btree_trans *trans,
 		     "unreachable inode:\n%s",
 		     (bch2_inode_unpacked_to_text(&buf, &inode),
 		      buf.buf)))
-		ret = bch2_reattach_inode(trans, &inode);
+		try(bch2_reattach_inode(trans, &inode));
 fsck_err:
 	return ret;
 }
@@ -1179,15 +1188,15 @@ fsck_err:
  */
 int bch2_check_unreachable_inodes(struct bch_fs *c)
 {
-	struct progress_indicator_state progress;
-	bch2_progress_init(&progress, c, BIT_ULL(BTREE_ID_inodes));
+	struct progress_indicator progress;
+	bch2_progress_init(&progress, __func__, c, BIT_ULL(BTREE_ID_inodes), 0);
 
 	CLASS(btree_trans, trans)(c);
 	return for_each_btree_key_commit(trans, iter, BTREE_ID_inodes,
 				POS_MIN,
 				BTREE_ITER_prefetch|BTREE_ITER_all_snapshots, k,
 				NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter) ?:
+		bch2_progress_update_iter(trans, &progress, &iter) ?:
 		check_unreachable_inode(trans, &iter, k);
 	}));
 }
@@ -1588,7 +1597,7 @@ static int check_dirent(struct btree_trans *trans, struct btree_iter *iter,
 		return 0;
 
 	if (dir->first_this_inode)
-		*hash_info = bch2_hash_info_init(c, &i->inode);
+		try(bch2_hash_info_init(c, &i->inode, hash_info));
 	dir->first_this_inode = false;
 
 	hash_info->cf_encoding = bch2_inode_casefold(c, &i->inode) ? c->cf_encoding : NULL;
@@ -1604,40 +1613,6 @@ static int check_dirent(struct btree_trans *trans, struct btree_iter *iter,
 
 	struct bkey_s_c_dirent d = bkey_s_c_to_dirent(k);
 
-	/* check casefold */
-	if (fsck_err_on(d.v->d_casefold != !!hash_info->cf_encoding,
-			trans, dirent_casefold_mismatch,
-			"dirent casefold does not match dir casefold\n%s",
-			(printbuf_reset(&buf),
-			 bch2_bkey_val_to_text(&buf, c, k),
-			 buf.buf))) {
-		subvol_inum dir_inum = { .subvol = d.v->d_type == DT_SUBVOL
-				? le32_to_cpu(d.v->d_parent_subvol)
-				: 0,
-		};
-		u64 target = d.v->d_type == DT_SUBVOL
-			? le32_to_cpu(d.v->d_child_subvol)
-			: le64_to_cpu(d.v->d_inum);
-		struct qstr name = bch2_dirent_get_name(d);
-
-		struct bkey_i_dirent *new_d =
-			errptr_try(bch2_dirent_create_key(trans, hash_info, dir_inum,
-					       d.v->d_type, &name, NULL, target));
-
-		new_d->k.p.inode	= d.k->p.inode;
-		new_d->k.p.snapshot	= d.k->p.snapshot;
-
-		struct btree_iter dup_iter = {};
-		return  bch2_hash_delete_at(trans,
-					    bch2_dirent_hash_desc, hash_info, iter,
-					    BTREE_UPDATE_internal_snapshot_node) ?:
-			bch2_str_hash_repair_key(trans, s,
-						 &bch2_dirent_hash_desc, hash_info,
-						 iter, bkey_i_to_s_c(&new_d->k_i),
-						 &dup_iter, bkey_s_c_null,
-						 need_second_pass);
-	}
-
 	if (d.v->d_type == DT_SUBVOL) {
 		try(check_dirent_to_subvol(trans, iter, d));
 	} else {
@@ -1705,17 +1680,17 @@ int bch2_check_dirents(struct bch_fs *c)
 	CLASS(snapshots_seen, s)();
 	CLASS(inode_walker, dir)();
 	CLASS(inode_walker, target)();
-	struct progress_indicator_state progress;
+	struct progress_indicator progress;
 	bool need_second_pass = false, did_second_pass = false;
 	int ret;
 again:
-	bch2_progress_init(&progress, c, BIT_ULL(BTREE_ID_dirents));
+	bch2_progress_init(&progress, __func__, c, BIT_ULL(BTREE_ID_dirents), 0);
 
 	ret = for_each_btree_key_commit(trans, iter, BTREE_ID_dirents,
 				POS(BCACHEFS_ROOT_INO, 0),
 				BTREE_ITER_prefetch|BTREE_ITER_all_snapshots, k,
 				NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-			progress_update_iter(trans, &progress, &iter) ?:
+			bch2_progress_update_iter(trans, &progress, &iter) ?:
 			check_dirent(trans, &iter, k, &hash_info, &dir, &target, &s,
 				     &need_second_pass);
 		})) ?:
@@ -1756,7 +1731,7 @@ static int check_xattr(struct btree_trans *trans, struct btree_iter *iter,
 		return 0;
 
 	if (inode->first_this_inode)
-		*hash_info = bch2_hash_info_init(c, &i->inode);
+		try(bch2_hash_info_init(c, &i->inode, hash_info));
 	inode->first_this_inode = false;
 
 	bool need_second_pass = false;
@@ -1773,8 +1748,8 @@ int bch2_check_xattrs(struct bch_fs *c)
 	CLASS(btree_trans, trans)(c);
 	CLASS(inode_walker, inode)();
 
-	struct progress_indicator_state progress;
-	bch2_progress_init(&progress, c, BIT_ULL(BTREE_ID_xattrs));
+	struct progress_indicator progress;
+	bch2_progress_init(&progress, __func__, c, BIT_ULL(BTREE_ID_xattrs), 0);
 
 	int ret = for_each_btree_key_commit(trans, iter, BTREE_ID_xattrs,
 			POS(BCACHEFS_ROOT_INO, 0),
@@ -1782,7 +1757,7 @@ int bch2_check_xattrs(struct bch_fs *c)
 			k,
 			NULL, NULL,
 			BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter) ?:
+		bch2_progress_update_iter(trans, &progress, &iter) ?:
 		check_xattr(trans, &iter, k, &hash_info, &inode);
 	}));
 	return ret;
@@ -1882,6 +1857,27 @@ int bch2_fix_reflink_p(struct bch_fs *c)
 			fix_reflink_p_key(trans, &iter, k));
 }
 
+/* translate to return code of fsck commad - man(8) fsck */
+int bch2_fs_fsck_errcode(struct bch_fs *c, struct printbuf *msg)
+{
+	int ret = 0;
+
+	if (test_bit(BCH_FS_errors_fixed, &c->flags)) {
+		prt_printf(msg, "%s: errors fixed\n", c->name);
+		ret |= 1;
+	}
+	if (test_bit(BCH_FS_error, &c->flags)) {
+		prt_printf(msg, "%s: still has errors\n", c->name);
+		ret |= 4;
+	}
+	if (test_bit(BCH_FS_emergency_ro, &c->flags)) {
+		prt_printf(msg, "%s: fatal error (went emergency read-only)\n", c->name);
+		ret |= 4;
+	}
+
+	return ret;
+}
+
 #ifndef NO_BCACHEFS_CHARDEV
 
 struct fsck_thread {
@@ -1901,26 +1897,21 @@ static int bch2_fsck_offline_thread_fn(struct thread_with_stdio *stdio)
 	struct fsck_thread *thr = container_of(stdio, struct fsck_thread, thr);
 	struct bch_fs *c = thr->c;
 
-	int ret = PTR_ERR_OR_ZERO(c);
+	errptr_try(c);
+
+	c->recovery_task = current;
+
+	int ret = bch2_fs_start(c);
+
+	CLASS(printbuf, buf)();
 	if (ret)
-		return ret;
-
-	thr->c->recovery_task = current;
-
-	ret = bch2_fs_start(thr->c);
+		prt_printf(&buf, "%s: error starting filesystem: %s\n", c->name, bch2_err_str(ret));
+	else
+		ret = bch2_fs_fsck_errcode(c, &buf);
 	if (ret)
-		goto err;
+		bch2_stdio_redirect_write(&stdio->stdio, false, buf.buf, buf.pos);
 
-	if (test_bit(BCH_FS_errors_fixed, &c->flags)) {
-		bch2_stdio_redirect_printf(&stdio->stdio, false, "%s: errors fixed\n", c->name);
-		ret |= 1;
-	}
-	if (test_bit(BCH_FS_error, &c->flags)) {
-		bch2_stdio_redirect_printf(&stdio->stdio, false, "%s: still has errors\n", c->name);
-		ret |= 4;
-	}
-err:
-	bch2_fs_stop(c);
+	bch2_fs_exit(c);
 	return ret;
 }
 
@@ -1929,15 +1920,18 @@ static const struct thread_with_stdio_ops bch2_offline_fsck_ops = {
 	.fn		= bch2_fsck_offline_thread_fn,
 };
 
+static int parse_mount_opts_user(char __user *optstr_user, struct bch_opts *opts)
+{
+	char *optstr __free(kfree) = errptr_try(strndup_user(optstr_user, 1 << 16));
+
+	return bch2_parse_mount_opts(NULL, opts, NULL, optstr, false);
+}
+
 long bch2_ioctl_fsck_offline(struct bch_ioctl_fsck_offline __user *user_arg)
 {
 	struct bch_ioctl_fsck_offline arg;
-	struct fsck_thread *thr = NULL;
-	darray_const_str devs = {};
-	long ret = 0;
 
-	if (copy_from_user(&arg, user_arg, sizeof(arg)))
-		return -EFAULT;
+	try(copy_from_user_errcode(&arg, user_arg, sizeof(arg)));
 
 	if (arg.flags)
 		return -EINVAL;
@@ -1945,42 +1939,30 @@ long bch2_ioctl_fsck_offline(struct bch_ioctl_fsck_offline __user *user_arg)
 	if (!capable(CAP_SYS_ADMIN))
 		return -EPERM;
 
+	struct bch_opts opts = bch2_opts_empty();
+	if (arg.opts)
+		try(parse_mount_opts_user((char __user *)(unsigned long) arg.opts, &opts));
+
+	CLASS(darray_const_str, devs)();
 	for (size_t i = 0; i < arg.nr_devs; i++) {
 		u64 dev_u64;
-		ret = copy_from_user_errcode(&dev_u64, &user_arg->devs[i], sizeof(u64));
-		if (ret)
-			goto err;
+		try(copy_from_user_errcode(&dev_u64, &user_arg->devs[i], sizeof(u64)));
 
-		char *dev_str = strndup_user((char __user *)(unsigned long) dev_u64, PATH_MAX);
-		ret = PTR_ERR_OR_ZERO(dev_str);
-		if (ret)
-			goto err;
+		char *dev_str =
+			errptr_try(strndup_user((char __user *)(unsigned long) dev_u64, PATH_MAX));
 
-		ret = darray_push(&devs, dev_str);
+		int ret = darray_push(&devs, dev_str);
 		if (ret) {
 			kfree(dev_str);
-			goto err;
+			return ret;
 		}
 	}
 
-	thr = kzalloc(sizeof(*thr), GFP_KERNEL);
-	if (!thr) {
-		ret = -ENOMEM;
-		goto err;
-	}
+	struct fsck_thread *thr = kzalloc(sizeof(*thr), GFP_KERNEL);
+	if (!thr)
+		return -ENOMEM;
 
-	thr->opts = bch2_opts_empty();
-
-	if (arg.opts) {
-		char *optstr = strndup_user((char __user *)(unsigned long) arg.opts, 1 << 16);
-		ret =   PTR_ERR_OR_ZERO(optstr) ?:
-			bch2_parse_mount_opts(NULL, &thr->opts, NULL, optstr, false);
-		if (!IS_ERR(optstr))
-			kfree(optstr);
-
-		if (ret)
-			goto err;
-	}
+	thr->opts = opts;
 
 	opt_set(thr->opts, stdio, (u64)(unsigned long)&thr->thr.stdio);
 	opt_set(thr->opts, read_only, 1);
@@ -1997,50 +1979,61 @@ long bch2_ioctl_fsck_offline(struct bch_ioctl_fsck_offline __user *user_arg)
 	    thr->c->opts.errors == BCH_ON_ERROR_panic)
 		thr->c->opts.errors = BCH_ON_ERROR_ro;
 
-	ret = __bch2_run_thread_with_stdio(&thr->thr);
-out:
-	darray_for_each(devs, i)
-		kfree(*i);
-	darray_exit(&devs);
+	int ret = __bch2_run_thread_with_stdio(&thr->thr);
+	if (ret < 0) {
+		if (thr)
+			bch2_fsck_thread_exit(&thr->thr);
+		pr_err("ret %s", bch2_err_str(ret));
+	}
 	return ret;
-err:
-	if (thr)
-		bch2_fsck_thread_exit(&thr->thr);
-	pr_err("ret %s", bch2_err_str(ret));
-	goto out;
 }
 
 static int bch2_fsck_online_thread_fn(struct thread_with_stdio *stdio)
 {
 	struct fsck_thread *thr = container_of(stdio, struct fsck_thread, thr);
 	struct bch_fs *c = thr->c;
+	CLASS(printbuf, buf)();
+	int ret = -EAGAIN;
 
-	c->stdio_filter = current;
-	c->stdio = &thr->thr.stdio;
+	if (mutex_trylock(&c->recovery.run_lock)) {
+		c->stdio_filter = current;
+		c->stdio = &thr->thr.stdio;
 
-	/*
-	 * XXX: can we figure out a way to do this without mucking with c->opts?
-	 */
-	unsigned old_fix_errors = c->opts.fix_errors;
-	if (opt_defined(thr->opts, fix_errors))
-		c->opts.fix_errors = thr->opts.fix_errors;
-	else
-		c->opts.fix_errors = FSCK_FIX_ask;
+		/*
+		 * XXX: can we figure out a way to do this without mucking with c->opts?
+		 */
+		unsigned old_fix_errors = c->opts.fix_errors;
+		if (opt_defined(thr->opts, fix_errors))
+			c->opts.fix_errors = thr->opts.fix_errors;
+		else
+			c->opts.fix_errors = FSCK_FIX_ask;
 
-	c->opts.fsck = true;
-	set_bit(BCH_FS_in_fsck, &c->flags);
+		c->opts.fsck = true;
+		set_bit(BCH_FS_in_fsck, &c->flags);
 
-	int ret = bch2_run_online_recovery_passes(c, ~0ULL);
+		ret = bch2_run_recovery_passes(c,
+			bch2_recovery_passes_match(PASS_FSCK) &
+			bch2_recovery_passes_match(PASS_ONLINE),
+			true) ?:
+			bch2_fs_fsck_errcode(c, &buf);
 
-	clear_bit(BCH_FS_in_fsck, &c->flags);
-	bch_err_fn(c, ret);
+		clear_bit(BCH_FS_in_fsck, &c->flags);
 
-	c->stdio = NULL;
-	c->stdio_filter = NULL;
-	c->opts.fix_errors = old_fix_errors;
+		c->stdio = NULL;
+		c->stdio_filter = NULL;
+		c->opts.fix_errors = old_fix_errors;
 
-	up(&c->recovery.run_lock);
+		mutex_unlock(&c->recovery.run_lock);
+	}
 	bch2_ro_ref_put(c);
+
+	if (ret < 0) {
+		prt_printf(&buf, "%s: error running recovery passes: %s\n", c->name, bch2_err_str(ret));
+		ret = 8;
+	}
+
+	if (buf.pos)
+		bch2_stdio_redirect_write(&stdio->stdio, false, buf.buf, buf.pos);
 	return ret;
 }
 
@@ -2051,51 +2044,32 @@ static const struct thread_with_stdio_ops bch2_online_fsck_ops = {
 
 long bch2_ioctl_fsck_online(struct bch_fs *c, struct bch_ioctl_fsck_online arg)
 {
-	struct fsck_thread *thr = NULL;
-	long ret = 0;
-
 	if (arg.flags)
 		return -EINVAL;
 
 	if (!capable(CAP_SYS_ADMIN))
 		return -EPERM;
 
+	struct bch_opts opts = bch2_opts_empty();
+	if (arg.opts)
+		try(parse_mount_opts_user((char __user *)(unsigned long) arg.opts, &opts));
+
 	if (!bch2_ro_ref_tryget(c))
 		return -EROFS;
 
-	if (down_trylock(&c->recovery.run_lock)) {
-		bch2_ro_ref_put(c);
-		return -EAGAIN;
-	}
-
-	thr = kzalloc(sizeof(*thr), GFP_KERNEL);
+	struct fsck_thread *thr = kzalloc(sizeof(*thr), GFP_KERNEL);
 	if (!thr) {
-		ret = -ENOMEM;
-		goto err;
+		bch2_ro_ref_put(c);
+		return -ENOMEM;
 	}
 
 	thr->c = c;
-	thr->opts = bch2_opts_empty();
+	thr->opts = opts;
 
-	if (arg.opts) {
-		char *optstr = strndup_user((char __user *)(unsigned long) arg.opts, 1 << 16);
-
-		ret =   PTR_ERR_OR_ZERO(optstr) ?:
-			bch2_parse_mount_opts(c, &thr->opts, NULL, optstr, false);
-		if (!IS_ERR(optstr))
-			kfree(optstr);
-
-		if (ret)
-			goto err;
-	}
-
-	ret = bch2_run_thread_with_stdio(&thr->thr, &bch2_online_fsck_ops);
-err:
+	int ret = bch2_run_thread_with_stdio(&thr->thr, &bch2_online_fsck_ops);
 	if (ret < 0) {
 		bch_err_fn(c, ret);
-		if (thr)
-			bch2_fsck_thread_exit(&thr->thr);
-		up(&c->recovery.run_lock);
+		bch2_fsck_thread_exit(&thr->thr);
 		bch2_ro_ref_put(c);
 	}
 	return ret;
